/**
 * Name: JmriDevice.java
 *
 * What:
 * This class extends JmriDevice to encompass decoder chains; thus is tightly coupled
 * to the decoder chain class.  What separates a decoder chain from a simple decoder
 * is that the chain has no hardware counterpart, but contains references to other decoders.
 * These references could be IOSpecs or JmriDevices - I need to decide which works best.
 * <p>
 * This class overrides the JmriDevice methods that deal directly with the embedded decoder
 * to account for the multiplicity of contained decoders.
 */
package cats.jmri;

import java.util.ArrayList;

import cats.layout.items.IOSpec;
import cats.layout.items.IOSpecChain;
import cats.layout.items.PtsEdge;

/**
 *  * This class extends JmriDevice to encompass decoder chains; thus is tightly coupled
 * to the decoder chain class.  What separates a decoder chain from a simple decoder
 * is that the chain has no hardware counterpart, but contains references to other decoders.
 * These references could be IOSpecs or JmriDevices - I need to decide which works best.
 * <p>
 * This class overrides the JmriDevice methods that deal directly with the embedded decoder
 * to account for the multiplicity of contained decoders.
 * 
  * <p>Title: CATS - Crandic Automated Traffic System</p>
 * <p>Description: A program for dispatching trains on Pat Lana's
 * Crandic model railroad.
 * <p>Copyright: Copyright (c) 2019, 2020</p>
 * <p>Company: </p>
 * @author Rodney Black
 * @version $Revision$
 */
public class JmriChainDevice extends JmriDevice {

	/**
	 * is true when instantiated and set to false on first usage. True means that the
	 * routes list should be expanded to include subtended route lists.
	 */
	private boolean NeedsExpansion;
	/**
	 * the ctor.  Either prefix+address or prefix+userName is required to uniquely identify
	 * the decoder.
	 * @param systemName is the System Name
	 * @param type is the JMRI device type
	 * @param user is the user name from the XML file
	 * 
	 */
	protected JmriChainDevice(final String systemName, final JmriDeviceType type, final String user) {
		super(systemName, type, user);
		NeedsExpansion = true;
	}

	@Override
	public void lockOutCommand() {
		ArrayList<IOSpec> decoders = ((IOSpecChain) MyDevice).getDecoders();
		for (IOSpec spec : decoders) {
			spec.lockOutCommand();
		}
		super.lockOutCommand();
	}
	
	@Override
	public void unlockOutCommand() {
		ArrayList<IOSpec> decoders = ((IOSpecChain) MyDevice).getDecoders();
		for (IOSpec spec : decoders) {
			spec.unlockOutCommand();
		}
		super.unlockOutCommand();
	}
	
	@Override
	public boolean isLockedOut() {
		ArrayList<IOSpec> decoders = ((IOSpecChain) MyDevice).getDecoders();
		for (IOSpec spec : decoders) {
			if (spec.isLockedOut()) {
				return true;
			}
		}
		return super.isLockedOut();
	}
	
	@Override
	public void addRouteInfo(final PtsEdge.RouteReference route, final boolean throwClose) {
		ArrayList<IOSpec> decoders = ((IOSpecChain) MyDevice).getDecoders();
		ArrayList<PtsEdge.RouteReference> points;
//		super.addRouteInfo(route, true);
//		super.addRouteInfo(route, false);
		if (throwClose) {  // "throw polarity" or the "undo" actions
			super.addRouteInfo(route, true);
			for (IOSpec spec : decoders) {
				if (spec.getExitFlag()) {
					points = spec.getDevice().getRoutes(!spec.getThrow());
					if (points != null) {
						for (PtsEdge.RouteReference pts : points) {
							super.addRouteInfo(pts, true);
						}
					}
				}
			}
		}
		else {
			super.addRouteInfo(route, false);
			for (IOSpec spec : decoders) {
				points = spec.getRegisteredUsers();
				if (points != null) {
					for (PtsEdge.RouteReference pts : points) {
						super.addRouteInfo(pts, false);
					}
				}
			}
		}
	}
	
	@Override
	public ArrayList<PtsEdge.RouteReference> getRoutes(final boolean throwClose) {
		if (NeedsExpansion) {
			expandRouteList(throwClose);
		}
		return super.getRoutes(throwClose);
	}
	
	/**
	 * is a method for adding the PtsEdges the decoders in the chain
	 * are involved in, to the chain's route list.
	 * @param throwClose if the polarity of the command is Throw, then throwClose is true;
	 * else it is false.
	 */
	private void expandRouteList(final boolean throwClose) {
		ArrayList<IOSpec> decoders = ((IOSpecChain) MyDevice).getDecoders();
		ArrayList<PtsEdge.RouteReference> points;
		if (throwClose) {
			for (IOSpec spec : decoders) {
				if (spec.getExitFlag()) {
					points = spec.getDevice().getRoutes(!spec.getThrow());
					if (points != null) {
						for (PtsEdge.RouteReference pts : points) {
							super.addRouteInfo(pts, true);
						}
					}
				}
			}
		}
		else {
			for (IOSpec spec : decoders) {
				points = spec.getRegisteredUsers();
				if (points != null) {
					for (PtsEdge.RouteReference pts : points) {
						super.addRouteInfo(pts, false);
					}
				}
			}
		}
		NeedsExpansion = false;
	}
}
/* @(#)JmriChainDevice.java */
