/* Name: PtsEdge.java
 *
 * What:
 *   This class is the container for all the information about one
 *   of a Section's Edges which has multiple routes, corresponding
 *   to switch points.
 */

package cats.layout.items;

import cats.common.Sides;
import cats.gui.DecoderInterlock;
import cats.gui.GridTile;
import cats.gui.LocalEnforcement;
import cats.layout.DecoderObserver;
import cats.rr_events.VerifyEvent;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.Enumeration;
import java.util.Vector;

/**
 * "is a" SecEdge which represents switch points - an edge on which
 * multiple tracks (routes) terminate.
 * <p>Title: CATS - Crandic Automated Traffic System</p>
 * <p>Description: A program for dispatching trains on Pat Lana's
 * Crandic model railroad.
 * <p>Copyright: Copyright (c) 2004, 2009, 2010, 2011, 2020</p>
 * <p>Company: </p>
 * @author Rodney Black
 * @version $Revision$
 */
public class PtsEdge
extends SecEdge {

	/**
	 * is a list of all PtsEdges on the layout.
	 */
	private static Vector<PtsEdge> PtsKeeper = new Vector<PtsEdge>();

	/**
	 * defines the number of SecEdges at the other end of routes that
	 * terminate on this SecEdge.
	 */
	public static final int MAX_ROUTES = Sides.EDGENAME.length;

	/**
	 * the following are indices into the Feedback BitSets, identifying the
	 * feedback properties of each route.
	 */
	
	/**
	 * set means that a route has been defined
	 */
	protected final int HAS_ROUTE = 0;
	
	/**
	 * set means that either in place or release feedback has been defined
	 */
	protected final int HAS_FEEDBACK = HAS_ROUTE + 1;
	
	/**
	 * set means in place feedback has been defined
	 */
	protected final int HAS_IN_PLACE = HAS_FEEDBACK + 1;
	
	/**
	 * set means the in place feedback has the same decoder address as the command
	 */
	protected final int COUPLED_IN_PLACE = HAS_IN_PLACE + 1;
	
	/**
	 * set means the in place feedback has a different decoder address than the command
	 */
	protected final int UNCOUPLED_IN_PLACE = COUPLED_IN_PLACE + 1;
	
	/**
	 * set means release feedback has been defined
	 */
	protected final int HAS_RELEASE = UNCOUPLED_IN_PLACE + 1;
	
	/**
	 * is the number of bits (indices) used by the above
	 */
	protected final int FEEDBACK_BITS = HAS_RELEASE + 1;
	
	/**
	 * the bitsets for each route
	 */
	protected final BitSet Feedback[] = {
			new BitSet(FEEDBACK_BITS),
			new BitSet(FEEDBACK_BITS),
			new BitSet(FEEDBACK_BITS),
			new BitSet(FEEDBACK_BITS)
	};
	
	/**
	 * the Signal protecting the Points from all directions.
	 */
	private OSSignal MySignal;

	/**
	 * the signal that precedes the points.
	 */
	private Signal PredSignal;

	/**
	 * the Signals protecting entry into the security plant from the frog
	 * end.
	 */
	private Signal FrogSignals[];

	/**
	 * the Signals which receive down stream Indications, from the next Signal
	 * in the direction of travel through the frog.  One of these is selected
	 * for determining the signal on the approach to the Points.
	 */
	private FrogSignal Feeders[];

	/**
	 * the speeds of each route through the frog.
	 */
	private int FrogSpeeds[] = new int[MAX_ROUTES];

	/**
	 * the state of the Block containing the OSEdge.
	 */
	protected Indication OSIndication;

	/**
	 * describes the routes through the points, how to sense the turnout
	 * position, how to throw the turnout, and how to set the lock light.
	 */
	protected SwitchPoints MyPoints;

	/**
	 * the Tracks terminating on this edge.  There can be a maximum of three
	 * tracks (one to each of the other edges), but the array contains four
	 * because each track is identified by the edge the other end terminates on.
	 */
	protected Track[] MyTracks = new Track[MAX_ROUTES];

	/**
	 * are the SignalWires for each route.
	 */
	protected SignalWire[] MyWires = new SignalWire[MAX_ROUTES];

	/**
	 * are the JMRI commands to move points
	 */
	protected final IOSpec[] Commands;

	/**
	 * the index of the currently selected track.  If it is the same as
	 * MyEdge, then the selection is not known.
	 */
	protected int CurrentTrk;

	/**
	 * is a flag indicating if the points have been locked in the field.  If true, then
	 * the local crew must unlock them before requests to move them will be
	 * honored.  This flag reflects that state of the field lock and is changed
	 * by the "Turn FieldLocked Light On" and "Turn Unlocked Light Off" layout responses.
	 */
	protected boolean FieldLocked = true;

	/**
	 * the Track with the Normal route through the points.
	 */
	protected int NormalRoute;
	
	/**
	 * is true if there is a listener for either the Lock or Unlock reply,
	 * indicating that CATS will be told when the field locks/unlocks
	 */
	private final boolean HasFieldLock;
	
	/**
	 * constructs a SecEdge with only its Edge identifier.
	 *
	 * @param edge identifies the side of the GridTile the SecEdge is on.
	 * @param shared describes the shared edge, if it is not adjacent.
	 *    A null value means it is adjacent.
	 * @param points describes the routes which terminate on the PtsEdge.
	 */
	public PtsEdge(final int edge, final Edge shared, final SwitchPoints points) {
		super(edge, shared);
		RouteInfo info;
		IOSpec decoder;
		IOSpec feedback;
		IOSpec[] localCtrl = new IOSpec[MAX_ROUTES];
		boolean toggle;
		DecoderObserver lockObserver = null;
		DecoderObserver unlockObserver = null;
		MyPoints = points;
		NormalRoute = MyPoints.getNormal();
		Commands = new IOSpec[MAX_ROUTES];
		if (points == null) { //they should always be not null.
			HasFieldLock = false;
		}
		else {

			// Though it may not be correct, start out with all fascia lock
			// indicators in the "locked" position.  It would be best to set
			// them to the current state, but the state cannot be read on some
			// systems (e.g. Loconet).
			if ( (decoder = MyPoints.getLockCmds(SwitchPoints.LOCKONCMD)) != null) {
				decoder.sendCommand();
			}

			// register an observer for the Unlock event
			if ( (decoder = MyPoints.getLockCmds(SwitchPoints.UNLOCKREPORT)) != null) {
				unlockObserver = new UnLockReceiver(decoder);
			}

			// register an observer for the Lock event
			if ( (decoder = MyPoints.getLockCmds(SwitchPoints.LOCKREPORT)) != null) {
				lockObserver = new LockReceiver(decoder);
			}
			HasFieldLock = (unlockObserver != null) || (lockObserver != null);
//			FieldLocked = !HasFieldLock;


			for (int align = 0; align < MAX_ROUTES; ++align) {
				if ( (info = MyPoints.getRouteList(align)) != null) {
					// register an observer for the local control requesting movement
					if ( (decoder = info.getRouteReport(RouteInfo.ROUTEREQUEST)) != null) {
						localCtrl[align] = decoder;
					}

					// register an observer for the points moving into position
					if ( (decoder = info.getRouteReport(RouteInfo.SELECTEDREPORT)) != null) {
						new InPlaceReceiver(align, decoder);
					}
					// register an observer for the points moving out of position
					if ( (decoder = info.getRouteReport(RouteInfo.NOTSELECTEDREPORT)) != null) {
						new ReleaseReceiver(align, decoder);
					}
				}
			}

			// Determine which switches are push buttons that just toggle between
			// selected routes
			for (int rte = 0; rte < MAX_ROUTES; ++rte) {
				toggle = false;
				if (localCtrl[rte] != null) {
					for (int r = rte + 1; r < MAX_ROUTES; ++r) {
						if ( (localCtrl[r] != null) && (localCtrl[rte].sameAs(localCtrl[r]))) {
							toggle = true;
							localCtrl[r] = null;
						}
					}
					if (toggle) {
						new ToggleReceiver(localCtrl[rte]);
					}
					else {
						new LocalReceiver(rte, localCtrl[rte]);
					}
				}
			}

			// collects the JmriDevices used for moving points and registers the PtsEdge
			// with the IOSpec representing the decoder.  Registration goes through the IOSPec
			// rather than directly to the device so that Chains can register their components.
			for (int side = 0; side < MAX_ROUTES; ++side) {
				if (((info = MyPoints.getRouteList(side)) != null) && ((decoder = info.getRouteCmd())!= null)) {
					Commands[side] = decoder;
					decoder.registerUser(new RouteReference(this, side));
				}
			}

			// establishes the kind of feedback for each route
			for (int path = 0; path < MAX_ROUTES; ++path) {
				if ((info = MyPoints.getRouteList(path)) != null) {
					Feedback[path].set(HAS_ROUTE);
					if ((feedback = info.getRouteReport(RouteInfo.SELECTEDREPORT)) != null) {
						Feedback[path].set(HAS_FEEDBACK);
						Feedback[path].set(HAS_IN_PLACE);
						if (Commands[path] == null) {
							Feedback[path].set(COUPLED_IN_PLACE);	// coupled Vs uncoupled makes no sense
						}
						else if (Commands[path].sameAs(feedback)) {
							Feedback[path].set(COUPLED_IN_PLACE);
						}
						else {
							Feedback[path].set(UNCOUPLED_IN_PLACE);
						}
					}
					if ((feedback = info.getRouteReport(RouteInfo.NOTSELECTEDREPORT)) != null) {
						Feedback[path].set(HAS_FEEDBACK);
						Feedback[path].set(HAS_RELEASE);
//						if ((decoder = info.getRouteCmd()) == null) {
//							Feedback[path].set(COUPLED_RELEASE);	// coupled Vs uncoupled makes no sense
//						}
//						else if (decoder.sameAs(feedback)) {
//							Feedback[path].set(COUPLED_RELEASE);
//						}
//						else {
//							Feedback[path].set(UNCOUPLED_RELEASE);
//						}
					}
					// this is needed to enable undo
					if ((Feedback[path].get(UNCOUPLED_IN_PLACE) || Feedback[path].get(HAS_RELEASE)) &&
							(info.getRouteCmd() != null)) {
						info.getRouteCmd().setExitFlag(true);
					}
				}
			}
		}
		MySignal = new OSSignal(this);
		FrogSignals = new Signal[MAX_ROUTES];
		Feeders = new FrogSignal[MAX_ROUTES];
		OSIndication = new Indication();
		OSIndication.setCondition(Block.TRK_RESERVED);
		PtsKeeper.add(this);
	}

	/*
	 * binds the edge to its mate.  It also converts the
	 * default speed to a real speed on each diverging route to "medium".
	 */
	public void bind() {
		super.bind();
		for (int trk = 0; trk < MyTracks.length; ++trk) {
			if ( (MyTracks[trk] != null) && (trk != NormalRoute) &&
					(MyTracks[trk].getSpeed() == Track.DEFAULT)) {
				MyTracks[trk].setSpeed(Track.MEDIUM);
			}
		}
		for (int rte = 0; rte < MAX_ROUTES; ++rte) {
			if (MyTracks[rte] != null) {
				Feeders[rte] = new FrogSignal(this);
			}
		}
		if (NormalRoute > -1) {
			Feeders[NormalRoute].setNormal();
		}
	}

	/**
	 * steps through the Tracks that have a termination on this Secedge,
	 * telling them which Block they are members of.
	 */
	public void propagateBlock(final Block block) {
		if (block != null) {
			for (int edge = 0; edge < MyTracks.length; ++edge) {
				if (MyTracks[edge] != null) {
					MyTracks[edge].setBlock(block);
				}
			}
			// if FieldLocked is false here, then there is no lock listener
//			if (HasFieldLock && ((block.getDiscipline() == Block.CTC) ||
//					(block.getDiscipline() == Block.DTC))) {
//				FieldLocked = true;
//			}
		}
	}

	/**
	 * remembers the Tracks that terminate on this SecEdge and returns true
	 * if the Track is the default route through the edge.
	 *
	 * @param trk is one of the Tracks terminating on the edge.
	 *
	 * @return true if the Track is on the Normal route.
	 */
	public boolean setDestination(final Track trk) {
		boolean normal = false;
		int other = trk.getTermination(this);
		if (other >= 0) {
			MyTracks[other] = trk;
			if (MyWires[other] == null) {
				MyWires[other] = trk;
			}
			if ( (NormalRoute == other) || (Destination == null)) {
				normal = true;
				Destination = trk;
				CurrentTrk = other;
				Wire = MyWires[other];
				//        movePoints(other);
				for (other = 0; other < MyTracks.length; ++other) {
					if ( (MyTracks[other] != null) && (CurrentTrk != other)) {
						MyTracks[other].switchEnd(this, false);
					}
				}
			}
			else {
				MyTracks[other].switchEnd(this, false);
			}
		}
		return normal;
	}

	/**
	 * remembers the wire for getting to the other side of the Section for
	 * linking Signals.
	 *
	 * @param link is the connection.
	 * @param which is the index of the other edge, for identifying the
	 * tracks in switch points
	 */
	public void setSignalWire(final SignalWire link, final int which) {
		MyWires[which] = link;
	}

	/**
	 * is called from a SecEdge across the Section, looking for the next signal
	 * in the direction of travel.  Since, this SecEdge is a PtsEdge, there
	 * will be a probe for each route.  All but the normal route will not
	 * have a next signal.
	 *
	 * @param rte is the Track of the path taken to get across the Section.
	 * It is needed by the PtsSec for identifying which of the routes is
	 * being probed.
	 *
	 * @param holder is a place to put the slowest speed found in the
	 * direction of travel.
	 *
	 * @return the next Signal encountered in the direction of travel.  It
	 * may be null.
	 */
	public Signal neighborProbe(final Track rte, final Indication holder) {
		int probe;
		for (probe = 0; probe < MAX_ROUTES; ++probe) {
			if (MyTracks[probe] == rte) {
				break;
			}
		}
		return Feeders[probe];
	}

	/**
	 * is called from a SecEdge in another Section, looking for the next signal
	 * in the direction of travel.  Since this SecEdge is not a Block and is
	 * not an OSEdge, it has no signal, so it crosses the Section on the
	 * normal route and passes the request to the next SecEdge in the
	 * direction of travel.
	 *
	 * @return the next Signal encountered in the direction of travel.  It
	 * may be null.
	 */
	public Signal signalProbe(final Indication holder) {
		return MySignal;
	}
	/**
	 * is called to locate the Signal's feeders.  It is called during
	 * initialization only.
	 */
	public void locateFeeder() {
		Indication speedHolder = new Indication(Track.DEFAULT, Block.TRK_RESERVED);
		Signal nextSignal;
		if (Joint != null) {
			if ( (nextSignal = Joint.signalProbe(speedHolder)) != null) {
				nextSignal.setPredecesor(MySignal);
			}
			ProtSpeed = speedHolder.getProtSpeed();
		}
		for (int route = 0; route < MAX_ROUTES; ++route) {
			if ( (MyTracks[route] != null) &&
					( (MyTracks[route].getDestination(this)) != null)) {
				speedHolder.setProtectedSpeed(MyTracks[route].getSpeed());
				if ( (nextSignal = MyWires[route].findSignal(this, speedHolder)) != null) {
					nextSignal.setPredecesor(Feeders[route]);
				}
				FrogSpeeds[route] = speedHolder.getProtSpeed();
			}
		}
	}

	/**
	 * is called to determine if local control is allowed or not.  To be allowed,
	 * the local lock must be unlocked and
	 * 
	 * if CTC/DTC
	 * - have track and time or out of service
	 * - have no reservation
	 * - not be occupied without T&T or OOS
	 * The Block.determineState() method checks for T&T and OOS before
	 * occupancy, so indirectly results in this conditional behavior.
	 * 
	 * if ABS/APB
	 * - have no reservation
	 * - T&T or OOS is optional and not necessary
	 *
	 * For a Chain, this should be tested on each PtsEdge that could be affected
	 * by issuing the any of the commands in the Chain.
	 * 
	 * @return true if the local control is active or false if it is
	 * ignored because the position of the points are locked.
	 */
	protected boolean isLocalControlAllowed() {
		boolean locked = false;
		if ((!HasFieldLock ||!FieldLocked) && !MyPoints.isRouteLocked()) {
			//      switch (TrkBlkState) {
			switch(MyBlock.determineState()) {
			case Block.TRK_AND_TIME:
			case Block.TRK_OOS:
				//      case Block.TRK_UNLOCKED:
				locked = true;
				break;

			case Block.TRK_IDLE:
				locked = (MyBlock.getDiscipline() != Block.CTC) && (MyBlock.getDiscipline() != Block.DTC);
				break;

			default:
			}
		}
		return locked;
	}

	/**
	 * is a predicate for determining if the dispatcher can move the points.  To be
	 * allowed to move the points,
	 * - local control must be locked (if it exists),
	 * - no reservation,
	 * - no occupancy
	 * - no T&T or OOS
	 * 
	 * No distinction is made for ABS/APB Vs CTC/DTC because dispatcher control of
	 * ABS/APB is undefined.  CATS allows it, so will treat moving points similar
	 * to under CTC/DTC.
	 * 
	 * Though a spur can not be lined by the dispatcher, the dispatcher may lock/unlock
	 * the points.  The spur points mat share a decoder address with OSEdge points.
	 * Consequently, this method must reside in PtsEdge, rather than OSEdge.
	 * 
	 * This test must be applied to every PtsEdge in a Chain that can be affected
	 * by commanding any decoder in the Chain.
	 * 
	 * @return true if the dispatcher can move the points
	 */
	protected boolean isDispatcherControlAllowed() {
		if ((!HasFieldLock || FieldLocked) &&	// Either the file lock listener is undefined or is set to lock
				(MyBlock.determineState() == Block.TRK_IDLE) &&	// handles TNT and OOS and occupancy
				!MyPoints.isRouteLocked()) {	// route reservation check
			return true;
		}
		return false;
	}

	/**
	 * is invoked to query all PtsEdges that share the decoder being asked
	 * to move its points.  If none object, then true is returned; otherwise,
	 * should any prohibit the action, false is returned.
	 * @param alignment is the side designating the track being requested
	 * @return true if all the PtsEdges sharing the decoder allow the action
	 * and false if any objects.
	 */
	protected boolean isFieldControlAllowed(final int alignment) {
		  ArrayList<RouteReference> points;
		  if (DecoderInterlock.TheInterlockType.getFlagValue() &&	// is the interlock check flag set?
				  (Commands[alignment] != null) &&					// is a command defined for the new route?
				  ((points = Commands[alignment].getRegisteredUsers()) != null)) {	// are other points sharing the decoder?
			  for (RouteReference pts : points) {
				  if (!pts.Edge.isLocalControlAllowed()) {
					  return false;
				  }
			  }
			  return true;		
		  }
		  return isLocalControlAllowed();
	}
	
	/**
	 * adjusts the fascia lock/unlock lights, based on the dispatcher
	 * granting/removing track and time and the previous value of
	 * track and time.
	 * @param state is the Track's state
	 */
	protected void adjustFasciaLights(final int state) {
		IOSpec cmd;
		if (!OSIndication.isTrackandTime() && (state == Block.TRK_AND_TIME)) {
			// Turn off the FieldLocked light and turn on the Unlocked light
			if ( (cmd = MyPoints.getLockCmds(SwitchPoints.LOCKOFFCMD)) != null) {
				cmd.sendCommand();
			}
			if ( (cmd = MyPoints.getLockCmds(SwitchPoints.LOCKONCMD)) != null) {
				cmd.sendUndoCommand();
			}
			if ( (cmd = MyPoints.getLockCmds(SwitchPoints.UNLOCKONCMD)) != null) {
				cmd.sendCommand();
			}
		}
		else if ( (OSIndication.isTrackandTime()) && (state != Block.TRK_AND_TIME)) {
			// Turn off the Unlocked light and turn on the FieldLocked light
			if ( (cmd = MyPoints.getLockCmds(SwitchPoints.UNLOCKOFFCMD)) != null) {
				cmd.sendCommand();
			}
			if ( (cmd = MyPoints.getLockCmds(SwitchPoints.UNLOCKONCMD)) != null) {
				cmd.sendUndoCommand();
			}
			if ( (cmd = MyPoints.getLockCmds(SwitchPoints.LOCKONCMD)) != null) {
				cmd.sendCommand();
			}
		}    
		OSIndication.setCondition(state);
	}

	/**
	 * receives the state of the Track.  For a plain Section, there is
	 * nothing to do.  It is here for several reasons.
	 * <ol>
	 * <li>it is invoked during initialization
	 * <li>it converts TRK_IDLE for ABS and APB to TRK_RESERVED
	 * </ol>
	 *
	 * @param state is the Track's state.
	 *
	 * @see cats.layout.items.Track
	 * @see cats.layout.items.Indication
	 */
	public void setState(int state) {
		//    Indication initial = new Indication();
		adjustFasciaLights(state);
		if ((state == Block.TRK_IDLE) && (MyBlock.getDiscipline() != Block.CTC) &&
				(MyBlock.getDiscipline() != Block.DTC)){
			state = Block.TRK_RESERVED;
		}
		super.setState(state);
		//    initial.setCondition(state);
		//    MySignal.trackState(initial);
		OSIndication.setCondition(state);
		MySignal.trackState(OSIndication);
		if (Feeders[CurrentTrk] != null) {
			//      Feeders[CurrentTrk].trackState(initial);
			Feeders[CurrentTrk].trackState(OSIndication);
		}
	}

	/**
	 * locks or unlocks the turnout decoders associated with the SecEdge.
	 * This does nothing for any edge but one with SwitchPoints.
	 * @param lock is true to lock the decoder commands and false to
	 * unlock them.
	 */
	public void setEdgeLock(boolean lock) {
		MyPoints.setRouteLock(CurrentTrk, lock);
	}

	/**
	 * used by the dispatcher to request that the points be moved.  Since the
	 * dispatcher controls only OS Points, this does nothing.
	 *
	 * @param trk identifies the requested aligned Track.
	 */
	public void setAlignment(final int trk) {
	}

	/**
	 * moves the points for the calling Track.  This moves the points.  For
	 * a turnout with feedback reporting, alignPoints would get
	 * invoked only after the message is received that the points moved.
	 *
	 * @param trk identifies the Track.
	 */
	protected void movePoints(final int trk) {
//		RouteInfo fromRte = MyPoints.getRouteList(CurrentTrk);
//		RouteInfo toRte = MyPoints.getRouteList(trk);
//		MySection.setSelTrack(MyTracks[trk]);
//		if ( ( (fromRte == null) ||
//				(fromRte.getRouteReport(RouteInfo.NOTSELECTEDREPORT) == null)) &&
//				( (toRte == null) ||
//						(toRte.getRouteReport(RouteInfo.SELECTEDREPORT) == null))) {
//			MyTracks[CurrentTrk].switchEnd(this, false);
//			alignPoints(trk);
//		}
//		else { // open the gap to show the rails are in motion.
////			MyTracks[CurrentTrk].switchEnd(this, false);
////			Destination = null;
////			EdgeTile.requestUpdate();
//			moveAlert(trk);
//		}
//		// The commands must be sent out last, in case the reports are
//		// listening for the commands
//		MyPoints.selectRoute(trk);
		ArrayList<RouteReference> points;
		MySection.setSelTrack(MyTracks[trk]);
		if (Commands[trk] != null)	{
			if ((points = Commands[trk].getRegisteredUsers()) != null) {	// are other points sharing the decoder?
				for (RouteReference pts : points) {
					pts.Edge.movePreparation(pts.TrackNumber);
				}
			}
			MyPoints.selectRoute(trk);
		}
		else {
			immediateMove(trk);
		}
	}
	
	/**
	 * this sets local data structure in preparation for sending a command
	 * to move the points.  The command could come from either this PtsEdge
	 * or a PtsEdge that shares a decoder.
	 * 
	 * @param track is the number of the track that the points will move to.
	 * It must not be -1.
	 */
	protected void movePreparation(final int track) {
		if (CurrentTrk != track) {
			if (Feedback[CurrentTrk].get(HAS_FEEDBACK)) {
				MyTracks[CurrentTrk].switchEnd(this, false);
				Destination = null;
				EdgeTile.requestUpdate();
			}
			else {
				immediateMove(track);
			}
		}
	}

	/**
	 * a method called when moving points (real with a command or virtual
	 * with no command defined) that do not have feedback.  It immediately
	 * adjusts all the linkages and the display for the new alignment.
	 * @param newRoute is the index of the track for the new alignment.
	 */
	private void immediateMove(final int newRoute) {
		MyTracks[CurrentTrk].switchEnd(this, false);
		alignPoints(newRoute);		
	}
	
	/**
	 * adjusts all the links when the points are moved.
	 *
	 * @param newRoute specifies which of the Tracks the points are aligned to.
	 */
	protected void alignPoints(final int newRoute) {
		if (newRoute != CurrentTrk) {
//			MyTracks[CurrentTrk].switchEnd(this, false);
			if (Feeders[CurrentTrk] != null) {
				Feeders[CurrentTrk].setRelay(null);
			}
			CurrentTrk = newRoute;
			Destination = MyTracks[CurrentTrk];
			Wire = MyWires[CurrentTrk];
			MySignal.setRelay(FrogSignals[CurrentTrk]);
			if (Feeders[CurrentTrk] != null) {
				Feeders[CurrentTrk].setRelay(PredSignal);
			}
			refreshState();
			//      MyBlock.resumeReservation();
		}
		MyTracks[CurrentTrk].switchEnd(this, true);
		EdgeTile.requestUpdate();
	}

	/**
	 * is a predicate called when a report is received of the points moving.
	 * It checks the various conditions and either allow or denies the movement.
	 * It checks only and takes no actions if denying the move.
	 * 
	 * This can be invoked from either a points released (moved out of place) or
	 * points in place report.  The difference is that trk for released should be
	 * CurrentTrk and trk will be the new alignment for an in place report.
	 * @param trk is the track the report is for.
	 * @return true if the movement should be accepted and false if it should be
	 * reversed.
	 */
	private boolean verifyMovement(final int trk) {
		// the first check is for things that mean the report should be accepted
		// unconditionally.
		if (!LocalEnforcement.TheEnforcementType.getFlagValue() ||	// override flag is turned off
				MyBlock.getTrkNTime() ||	// Track and Time is on
				MyBlock.getOOS() ||			// Out of Service is on
				(Destination == null) ||	// points are in motion
				(trk == CurrentTrk)) {		// response due to another set of points reversing action
			return true;
		}
		
		// here are reasons why the movement should be denied.
//		if (((CurrentTrk >= 0) && MyPoints.isRouteLocked(CurrentTrk)) || 
//				((trk > 0) && MyPoints.isRouteLocked(trk))) {	// a route reservation is in existence
		if (MyPoints.isRouteLocked()) {
			return false;
		}
		
		// Discipline specific reasons for not allowing the points to move.
		if ((MyBlock.getDiscipline() == Block.CTC) || (MyBlock.getDiscipline() == Block.DTC)) {
			if (Destination != null) {	// points are not in motion, so field initiated
				return false;
			}
		}
		return true;
	}
	
	/**
	 * is invoked when points move into place to verify that the
	 * movement is expected.  If not, a reversing command is sent.
	 * This enforces that the crew cannot move the points, unless
	 * permitted by the dispatcher.
	 * 
	 * The action for the VerifyEvent will insure that a decoder has been defined
	 * to restore the points.
	 * 
	 * @param trk is the index of the track that the points are (have
	 * been commanded) to move to.
	 * @return true if the movement report is accepted and false if not and
	 * a reversing move has been scheduled
	 */
	private boolean verifyPointsinPlace(final int trk) {
		//		if (LocalEnforcement.TheEnforcementType.getFlagValue() &&
		//				((MyBlock.getDiscipline() == Block.CTC) || (MyBlock.getDiscipline() == Block.DTC)) &&
		//				!MyBlock.getTrkNTime() && !MyBlock.getOOS()) {
		//			//      MyPoints.verifyMovement(trk);
		VerifyEvent request;
		if (!verifyMovement(trk) &&		// local conditions prohibit the movement
				(CurrentTrk >= 0)) {	// movement is not expected
//			System.out.println("Reverse");
//			if (Feedback[trk].get(UNCOUPLED_IN_PLACE)) {
//				request = new VerifyEvent(Commands[CurrentTrk], false);
//				request.queUp();
//			}
			request = new VerifyEvent(Commands[CurrentTrk], true);
			request.queUp();
			return false;
		}
		return true;
	}

	/**
	 * is invoked when points move out of place to verify that the
	 * movement is expected.  If not, a countermanding command is sent.
	 * This enforces that the crew cannot move the points, unless
	 * permitted by the dispatcher.
	 * 
	 * The action for the VerifyEvent will insure that a decoder has been defined
	 * to restore the points.
	 * 
	 * @param trk is the index of the track that the points are (have
	 * been commanded) to move to.
	 * @return true if the movement report is accepted and false if not and
	 * a reversing move has been scheduled
	 */
	private boolean verifyPointsRelease(final int trk) {
//		if (LocalEnforcement.TheEnforcementType.getFlagValue() &&
//				((MyBlock.getDiscipline() == Block.CTC) || (MyBlock.getDiscipline() == Block.DTC)) &&
//				!MyBlock.getTrkNTime() && !MyBlock.getOOS()) {
//			//      MyPoints.verifyNoMovement(trk);
		VerifyEvent request;
		if (!verifyMovement(-1)) {		// local conditions prohibit the release
//			System.out.println("Reverse");
			if (Feedback[trk].get(HAS_RELEASE)) {
				request = new VerifyEvent(Commands[CurrentTrk], false);
				request.queUp();
			}
			request = new VerifyEvent(Commands[CurrentTrk], true);
			request.queUp();
			return false;
		}
		return true;
	}

	/**
	 * Registers the Signal that precedes the Points Signal.
	 *
	 * @param pts is the Signal a train encounters before it approaches
	 * the OS security element from the points end.
	 */
	public void setPtsSignal(final Signal pts) {
		PredSignal = pts;
		Feeders[NormalRoute].initialSignal(pts);
	}

	/**
	 * Registers a Signal that requests Indications for a route through the
	 * frog.  The requesting Signal is protecting entry into the route through
	 * the frog.
	 *
	 * @param reqSig is the Signal requesting to be fed Indications.
	 * @param frogSig is the Signal that received the request.
	 */
	public void setFrogSignal(final Signal reqSig, final FrogSignal frogSig) {
		int rte;
		Indication ind;
		for (rte = 0; rte < MAX_ROUTES; ++rte) {
			if (Feeders[rte] == frogSig) {
				break;
			}
		}
		FrogSignals[rte] = reqSig;
		if (rte == NormalRoute) {
			MySignal.initialSignal(reqSig);
		}
		else {
			ind = new Indication();
			ind.setCondition(Block.TRK_FOULED);
			FrogSignals[rte].nextSignal(ind);

		}
	}

	/**
	 * computes a new state based on the Points alignment and Track state.
	 * Since MySignal feeds the Signal protecting entry into the OS
	 * security plant via the current algnment, that Frog Signal can be
	 * ignored.  The Feeders are also updated, but because only one is active,
	 * not as much care is needed in copying the Indication.
	 */
	protected void refreshState() {
		Indication ind;
		OSIndication.setProtectedSpeed(ProtSpeed);
		MySignal.trackState(OSIndication);
		for (int rte = 0; rte < MAX_ROUTES; ++rte) {
			if (FrogSignals[rte] != null) {
				if (CurrentTrk != rte) {
					ind = new Indication(OSIndication);
					//          ind.setNoRoute();
					ind.setCondition(Block.TRK_FOULED);
					FrogSignals[rte].nextSignal(ind);
				}
			}
			if (Feeders[rte] != null) {
				OSIndication.setProtectedSpeed(FrogSpeeds[rte]);
				Feeders[rte].trackState(OSIndication);
			}
		}
	}

	/**
	 * steps through the list of PtsEdges, setting them all to the 
	 * last known alignment.
	 */
	public static void restoreAlignment() {
		PtsEdge turnOut;
		for (Enumeration<PtsEdge> e = PtsKeeper.elements(); e.hasMoreElements(); ) {
			turnOut = e.nextElement();
			turnOut.MyPoints.restorePoints();
		}
	}

	/**
	 * steps through the list of PtsEdges, setting them all to the a
	 * particular state.  The intent of this method is to test the turnouts
	 * on a layout.  If tested for throw, then closed, then
	 * if the layout is refreshed, CATS will know the positions
	 * of all tunrouts.
	 * 
	 * @param state is true to throw them all or false to close
	 * them all.
	 */
	public static void testThrown(final boolean state) {
		PtsEdge turnOut;
		for (Enumeration<PtsEdge> e = PtsKeeper.elements(); e.hasMoreElements(); ) {
			turnOut = e.nextElement();
			turnOut.testTurnout(state);
		}
	}

	/**
	 * positions all the turnout decoders controlled by the
	 * PtsEdge to a particular state.
	 * 
	 * @param state is the state being tested.  It is true
	 * for thrown and false for closed.
	 */
	public void testTurnout(final boolean state) {
		if (!MyBlock.getOccupied() && (MyPoints != null)) {
			MyPoints.testRoutes(state);
		}
	}

	/**
	 * positions all turnouts to the normal (straight thru) route.  If
	 * the turnout has feedback, stimulate the method that moves the
	 * points.  If it doesn't, then stimulate the method that moves
	 * the points and changes the graphics.
	 * <p>
	 * This does not move the points if the block is occupied
	 *
	 */
	public static void setAllNormal() {
		PtsEdge turnOut;
		for (Enumeration<PtsEdge> e = PtsKeeper.elements(); e.hasMoreElements(); ) {
			turnOut = e.nextElement();
			if (!turnOut.MyBlock.getOccupied() &&(turnOut.MyPoints != null)) {
				if (turnOut.MyPoints.hasFeedback()) {
					turnOut.MyPoints.selectRoute(turnOut.NormalRoute);
				}
				else {
					turnOut.setAlignment(turnOut.NormalRoute);
				}
			}
		}
	}

	/**
	 * steps through the list of PtsEdges, setting all the Lock lights.  The
	 * intent of this method is to initialize the lights on a layout that
	 * does not report state when the system is powered up or to test
	 * all the lights.
	 * <p>
	 * The actions here are a little confusing because there could be 2
	 * ways to turn off the existing light.  For example, to turn on
	 * the Lock light, the Unlock light should be turned off.  If the
	 * command to turn it off (UNLOCKOFF) is defined, it is sent.
	 * Alternatively, if the UNLOCKON has a cancel operation, it is sent.
	 * @param locked is true if the light is to show "locked"; false, if
	 * it is to show "unlocked".
	 */
	public static void testLockLights(final boolean locked) {
		PtsEdge turnOut;
		IOSpec decoder;
		for (Enumeration<PtsEdge> e = PtsKeeper.elements(); e.hasMoreElements(); ) {
			turnOut = e.nextElement();
			if (locked) {
				if ( (decoder = turnOut.MyPoints.getLockCmds(SwitchPoints.UNLOCKOFFCMD)) != null) {
					decoder.sendCommand();
				}
				if ( (decoder = turnOut.MyPoints.getLockCmds(SwitchPoints.UNLOCKONCMD)) != null) {
					decoder.sendUndoCommand();
				}
				if ( (decoder = turnOut.MyPoints.getLockCmds(SwitchPoints.LOCKONCMD)) != null) {
					decoder.sendCommand();
				}
			}
			else {
				if ( (decoder = turnOut.MyPoints.getLockCmds(SwitchPoints.LOCKOFFCMD)) != null) {
					decoder.sendCommand();
				}
				if ( (decoder = turnOut.MyPoints.getLockCmds(SwitchPoints.LOCKONCMD)) != null) {
					decoder.sendUndoCommand();
				}
				if ( (decoder = turnOut.MyPoints.getLockCmds(SwitchPoints.UNLOCKONCMD)) != null) {
					decoder.sendCommand();
				}
			}
		}
	}

	/**
	 * is an inner class for receiving requests for local movement of the points.
	 * It has a one-to-one relationship with a single alignment of the points.
	 */
	private class LocalReceiver
	implements DecoderObserver {

		/**
		 * is the Track being controlled.
		 */
		private final int TrkControlled;

		/**
		 * is the constructor.
		 *
		 * @param c is the track being controlled.
		 * @param decoder is the IOSpec describing the messages that the receiver
		 * is looking for.
		 */
		public LocalReceiver(final int c, final IOSpec decoder) {
			TrkControlled = c;
			decoder.registerListener(this);
		}

		/**
		 * is the interface through which the MsgFilter delivers the RREvent.
		 * This routine probably should tell the enclosing Section that the
		 * alignment has changed.  Not doing so means that the dispatcher may
		 * have to click multiple times to make a change.
		 */
		public void acceptIOEvent() {
			if (isFieldControlAllowed(TrkControlled) && (TrkControlled != CurrentTrk)) {
				movePoints(TrkControlled);
				GridTile.doUpdates();
			}
		}
	}

	/**
	 * is an inner class for receiving requests for local movement of the points.
	 * It is a toggle report so that every time it is received, the points move.
	 * This can be confusing to test.  A momentary push button will probably send a make,
	 * then a break, so JMRI will send two reports.  If this code reacted to both,
	 * it would effectively do nothing.  Thus, it listens to only one.
	 */
	private class ToggleReceiver
	implements DecoderObserver {

		/**
		 * is the constructor.
		 *
		 * @param decoder is the IOSpec describing the messages that the receiver
		 * is looking for.
		 */
		public ToggleReceiver(final IOSpec decoder) {
			decoder.registerListener(this);
		}

		/**
		 * is the interface through which the MsgFilter delivers the RREvent.
		 * This routine probably should tell the enclosing Section that the
		 * alignment has changed.  Not doing so means that the dispatcher may
		 * have to click multiple times to make a change.
		 */
		public void acceptIOEvent() {
			int newRoute = CurrentTrk;
			do {
				if ( (++newRoute) == MyTracks.length) {
					newRoute = 0;
				}
			}
			while (MyTracks[newRoute] == null);
			if (isFieldControlAllowed(newRoute)) {
				movePoints(newRoute);
				GridTile.doUpdates();
			}
		}
	}

	/**
	 * is an inner class for receiving reports that the points are in position.
	 * It has a one-to-one relationship with a single alignment of the points.
	 */
	private abstract class AlignmentReceiver
	implements DecoderObserver {

		/**
		 * is the Track that is aligned.
		 */
		protected int TrkControlled;

		/**
		 * is the constructor.
		 *
		 * @param c is the track being aligned.
		 * @param decoder is the IOSpec describing the messages that the receiver
		 * is looking for.
		 */
		public AlignmentReceiver(final int c, final IOSpec decoder) {
			TrkControlled = c;
			decoder.registerListener(this);
		}

		/*
		 * is the interface through which the MsgFilter delivers the RREvent.
		 */
//		public void acceptIOEvent() {
////			MyBlock.clrRoute();
//			doAlignment();
////			GridTile.doUpdates();
//		}

		/**
		 * is the routine that handles the report.
		 */
//		protected abstract void doAlignment();
		public abstract void acceptIOEvent();
	}

	/**
	 * is an inner class that reports on the points being moved into position.
	 */
	private class InPlaceReceiver
	extends AlignmentReceiver {

		/**
		 * is the constructor.
		 *
		 * @param c is the track being aligned.
		 * @param decoder is the IOSpec describing the messages that the receiver
		 * is looking for.
		 */

		public InPlaceReceiver(final int c,final IOSpec decoder) {
			super(c, decoder);
		}

		/**
		 * tells the PtsEdge that the points are aligned
		 */
//		protected void doAlignment() {
		public void acceptIOEvent() {
			MyTracks[CurrentTrk].switchEnd(PtsEdge.this, false);
			if (verifyPointsinPlace(TrkControlled)) {
				if (CurrentTrk != TrkControlled) {
					MyBlock.clrRoute();
				}
				alignPoints(TrkControlled);
				MySection.setSelTrack(MyTracks[TrkControlled]);
			}
			GridTile.doUpdates();			
		}
	}

	/**
	 * is an inner class that reports on the points being moved out of position.
	 */
	private class ReleaseReceiver
	extends AlignmentReceiver {

		/**
		 * is the constructor.
		 *
		 * @param c is the track being aligned.
		 * @param decoder is the IOSpec describing the messages that the receiver
		 * is looking for.
		 */

		public ReleaseReceiver(final int c, final IOSpec decoder) {
			super(c, decoder);
		}

		/**
		 * tells the PtsEdge that the points are not aligned
		 */
//		protected void doAlignment() {
		public void acceptIOEvent() {
			int otherTrk;  // This cannot be computed when the class is instantiated because
			// MyTracks may not have been filled in.  It is the index into
			// MyTracks of the other route.  This does not work for a three-way
			// turnout.

			if (verifyPointsRelease(TrkControlled)) {
				if (CurrentTrk != TrkControlled) {
					MyBlock.clrRoute();
				}
				for (otherTrk = 0; otherTrk < MyTracks.length; ++otherTrk) {
					if ((MyTracks[otherTrk] != null) && (otherTrk != TrkControlled)) {
						break;
					}
				}
				if (!Feedback[otherTrk].get(HAS_IN_PLACE)) {
					//				verifyPointsRelease(TrkControlled);
					alignPoints(otherTrk);
					MySection.setSelTrack(MyTracks[otherTrk]);
				}
				else {
					// show the points no longer aligned for TrkControlled
					MyTracks[CurrentTrk].switchEnd(PtsEdge.this, false);
				}
			}
			else {
				MyTracks[CurrentTrk].switchEnd(PtsEdge.this, false);
			}
			GridTile.doUpdates();
		}
	}

	/**
	 * is an inner class for receiving reports that the manual controls on
	 * the switch machine have been unlocked.
	 */
	private class UnLockReceiver
	implements DecoderObserver {

		/**
		 * is the constructor.
		 *
		 * @param decoder is the IOSpec describing the messages that the receiver
		 * is looking for.
		 */
		public UnLockReceiver(final IOSpec decoder) {
			decoder.registerListener(this);
		}

		/*
		 * is the interface through which the MsgFilter delivers the RREvent.
		 */
		public void acceptIOEvent() {
			FieldLocked = false;
			GridTile.doUpdates();
		}
	}

	/**
	 * is an inner class for receiving reports that the manual controls on
	 * the switch machine have been locked.
	 */
	private class LockReceiver
	implements DecoderObserver {

		/**
		 * is the constructor.
		 *
		 * @param decoder is the IOSpec describing the messages that the receiver
		 * is looking for.
		 */
		public LockReceiver(final IOSpec decoder) {
			decoder.registerListener(this);
		}

		/*
		 * is the interface through which the MsgFilter delivers the RREvent.
		 */
		public void acceptIOEvent() {
			FieldLocked = true;
			GridTile.doUpdates();
		}
	}

	/**
	 * is a container for describing a route- the PtsEdge and the track number.
	 */
	public class RouteReference {
		
		/**
		 * the PtsEdge
		 */
		public final PtsEdge Edge;
		
		/**
		 * the track number
		 */
		public final int TrackNumber;
		
		/**
		 * the ctor
		 * @param edge is the PtsEdge
		 * @param track is the track number
		 */
		RouteReference(final PtsEdge edge, final int track) {
			Edge = edge;
			TrackNumber = track;
		}
		
	}
	
	static org.apache.log4j.Logger log = org.apache.log4j.Logger.getLogger(
			PtsEdge.class.getName());
}
/* @(#)PtsEdge.java */
